# Copyright 2023-2025 Jeongmin Kang, jarvisNim @ GitHub
# See LICENSE for details.

__author__ = "jarvisNim in GitHub"
__version__ = "0.0.5"


from .base import (
    get_stock_history_by_fmp,
    get_stock_history_by_yfinance,
    trend_detector,
    trend_detector_for_series,
    get_working_day_before,
)

from .tech import (
    get_tech_types,
    get_tech_hist,
    get_tech_fin,    
    make_tech_plot,
    sma,
    ema,
    macd,
    adx,
    psar,
    ichmoku,
    rsi,
    stoch,
    roc,
    cci,
    willr,
    ao,
    stochrsi,
    ppo,
    obv,
    pvt,
    pvi,
    cmf,
    vwap,
    adosc,
    mfi,
    kvo,
    nvi,
    atr,
    bbands,
    donchian,
    kc,
    rvi,
)

from .strategy import (
    sma_strategy,
    timing_strategy,
    get_vb_signals,
    show_vb_stategy_result,
    volatility_bollinger_strategy,
    get_reversal_signals,
    show_reversal_stategy_result,
    reversal_strategy,
    trend_following_strategy,
    control_chart_strategy,
    vb_genericAlgo_strategy,
    vb_genericAlgo2_strategy,
    gaSellHoldBuy_strategy,
)

